# A "provider" is a Python class which provides a profile for a given
# language. Instead of supplying the language's checks as a YAML file,
# the developer can supply a provider which knows how to generate
# the checks at runtime. They have a single method, `.fill`, which
# takes a language definition and adds any relevant checks to its
# "shaperglot_checks" list.
from .arabic import ArabicProvider
from .african_latin import AfricanLatinProvider

PROVIDERS = [
    ArabicProvider,
    AfricanLatinProvider,
]


from functools import lru_cache

from youseedee import ucd_data

from shaperglot.checks.orthographies import OrthographiesCheck

from .common import shaping_input_schema, ShaperglotCheck, check_schema


@lru_cache(maxsize=None)
def _simple_mark_check(codepoint):
    return ucd_data(codepoint).get("General_Category") == "Mn"


class NoOrphanedMarksCheck(ShaperglotCheck):
    name = "no_orphaned_marks"
    schema = check_schema(
        {
            "input": shaping_input_schema,
        }
    )

    def describe(self):
        return f"that, when {self.input.describe()}, no marks are left unattached"

    def execute(self, checker):
        if not self.input.text:
            return
        buffer = self.input.shape(checker)
        dotted_circle_glyph = checker.cmap.get(0x25CC)

        # GDEF may be wrong, don't trust it.
        passed = True
        previous = None
        # pylint: disable=C0103
        for ix, (info, pos) in enumerate(
            zip(buffer.glyph_infos, buffer.glyph_positions)
        ):
            glyphname = checker.glyphorder[buffer.glyph_infos[ix].codepoint]
            # Is this a mark glyph?
            if info.codepoint == 0:
                passed = False
                message = "Shaper produced a .notdef"
                # Generally this message happens because we're missing
                # base/mark glyphs, so we will heard about this already,
                # *except* if no exemplars are defined, when it would be
                # good to hear specifically what glyphs we tried to test here.
                if not checker.lang.get("exemplarChars"):
                    message += " when shaping " + self.input.describe()

                checker.results.fail(
                    check_name="no-orphaned-marks",
                    result_code="notdef-produced",
                    message=message,
                    context={"text": self.input.check_yaml},
                )
                break
            if _simple_mark_check(checker.codepoint_for(glyphname)):
                # Was the previous glyph dotted circle?
                if previous and previous == dotted_circle_glyph:
                    passed = False
                    checker.results.fail(
                        check_name="no-orphaned-marks",
                        result_code="dotted-circle-produced",
                        message="Shaper produced a dotted circle when shaping "
                        + self.input.describe(),
                        context={"text": self.input.check_yaml},
                    )
                elif pos.x_offset == 0 and pos.y_offset == 0:  # Suspicious
                    passed = False
                    checker.results.fail(
                        check_name="no-orphaned-marks",
                        result_code="orphaned-mark",
                        message=f"Shaper didn't attach {glyphname} to {previous}",
                        context={
                            "text": self.input.check_yaml,
                            "mark": glyphname,
                            "base": previous,
                        },
                    )
            previous = glyphname
        if passed:
            checker.results.okay(
                check_name="no-orphaned-marks",
                message="No unattached mark glyphs were produced "
                + self.input.describe(),
                context={"text": self.input.check_yaml},
            )


class NoOrphanedMarksInOrthographiesCheck(NoOrphanedMarksCheck):
    name = "no_orphaned_marks_in_orthographies"

    def __init__(self, lang):
        super().__init__({
            "input": {
                "text":" ".join(OrthographiesCheck(lang).bases)
            }
        })



import argparse
import sys
from textwrap import fill
import os
import re

from shaperglot.checker import Checker
from shaperglot.languages import Languages


def describe(options):
    """Describe the checks shaperglot will perform to determine support for a given language"""
    langs = Languages()
    if options.lang not in langs:
        maybe = langs.disambiguate(options.lang)
        if len(maybe) == 1:
            lang = langs[maybe[0]]
            print(f"Assuming you meant {maybe[0]} ({lang['full_name']}).")
        elif len(maybe) > 1:
            print(f"Language '{options.lang}' not known", end="")
            print("; try one of: " + ", ".join(maybe))
            return
        else:
            print(f"Language '{options.lang}' not known", end="")
            print("")
            return
    else:
        lang = langs[options.lang]
    print(f"To test for {lang['name']} support, shaperglot will:")
    for shaperglot_check in lang.get("shaperglot_checks", []):
        print(
            fill(
                "ensure " + shaperglot_check.describe(),
                initial_indent=" * ",
                subsequent_indent="   ",
                width=os.get_terminal_size()[0] - 2,
            )
        )


def check(options):
    """Check a particular language or languages are supported"""
    checker = Checker(options.font)
    langs = Languages()
    for lang in options.lang:
        if lang not in langs:
            print(f"Language '{options.lang}' not known")
            continue

        results = checker.check(langs[lang])

        if results.is_unknown:
            print(f"Cannot determine whether font supports language '{lang}'")
        elif results.is_success:
            print(f"Font supports language '{lang}'")
        else:
            print(f"Font does not fully support language '{lang}'")

        if options.verbose and options.verbose > 1:
            for message in results:
                print(f" * {message}")
        elif options.verbose or not results.is_success:
            for message in results.fails:
                print(f" * {message}")


def report(options):
    """Report which languages are supported by the given font"""
    checker = Checker(options.font)
    langs = Languages()
    messages = []
    supported = []
    unsupported = []

    for lang in sorted(langs.keys()):
        if options.filter and not re.search(options.filter, lang):
            continue
        results = checker.check(langs[lang])

        if results.is_unknown:
            continue

        if results.is_success:
            supported.append(langs[lang]['name'])
            print(f"Font supports language '{lang}' ({langs[lang]['name']})")
        else:
            unsupported.append(langs[lang]['name'])
            print(
                f"Font does not fully support language '{lang}' ({langs[lang]['name']})"
            )
        messages.extend(results)
        if options.verbose and options.verbose > 1:
            for status, message in results:
                print(f" * {status.value}: {message}")

    # Collate a useful fixing guide
    print("\n== Summary ==\n")
    print(f"* {len(supported)+len(unsupported)} languages checked")
    if supported:
        print(f"* {len(supported)} languages supported")
    if not options.verbose:
        return
    if unsupported:
        print(
            fill(
                "* Unsupported languages: " + ", ".join(unsupported),
                subsequent_indent=" " * 25,
                width=os.get_terminal_size()[0] - 2,
            )
        )
        print("\nTo add support:")
    missing_bases = set()
    missing_marks = set()
    missing_anchors = set()
    for msg in messages:
        if msg.result_code == "bases-missing":
            missing_bases |= set(msg.context["glyphs"])
        if msg.result_code == "marks-missing":
            missing_marks |= set(msg.context["glyphs"])
        if msg.result_code == "orphaned-mark":
            missing_anchors.add((msg.context["base"], msg.context["mark"]))
    if missing_marks:
        print(
            fill(
                " * Add mark glyphs: "
                + ", ".join(["\u25cc" + x for x in sorted(missing_marks)]),
                subsequent_indent="    ",
                width=os.get_terminal_size()[0] - 2,
            )
        )
    if missing_bases:
        print(
            fill(
                " * Support characters: " + ", ".join(sorted(missing_bases)),
                subsequent_indent="    ",
                width=os.get_terminal_size()[0] - 2,
            )
        )
    if missing_anchors:
        print(
            fill(
                " * Add anchor attachments: "
                + ", ".join(
                    [base + '/' + mark for base, mark in sorted(missing_anchors)]
                ),
                subsequent_indent="    ",
                width=os.get_terminal_size()[0] - 2,
            )
        )


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    parser = argparse.ArgumentParser(
        description="Check a font file's language coverage"
    )
    subparsers = parser.add_subparsers(help='sub-commands')

    parser_describe = subparsers.add_parser('describe', help=describe.__doc__)
    parser_describe.add_argument(
        'lang', metavar='LANG', help='an ISO639-3 language code'
    )
    parser_describe.set_defaults(func=describe)

    parser_check = subparsers.add_parser('check', help=check.__doc__)
    parser_check.add_argument('--verbose', '-v', action='count')
    parser_check.add_argument('font', metavar='FONT', help='the font file')
    parser_check.add_argument(
        'lang', metavar='LANG', help='one or more ISO639-3 language codes', nargs="+"
    )
    parser_check.set_defaults(func=check)

    parser_report = subparsers.add_parser('report', help=report.__doc__)
    parser_report.add_argument('font', metavar='FONT', help='the font file')
    parser_report.add_argument('--verbose', '-v', action='count')
    parser_report.add_argument(
        '--filter', type=str, help="Regular expression to filter languages"
    )
    parser_report.set_defaults(func=report)

    options = parser.parse_args(args)
    if not hasattr(options, "func"):
        parser.print_help()
        sys.exit(1)
    options.func(options)


if __name__ == '__main__':  # pragma: no cover
    main()



# The main thing this module provides is the Languages class, which is a
# superset of the LoadLanguages() object from gflanguages. When a language is
# requested, we copy the gflanguages protobuf object into a plain Python
# dictionary, and then add the shaperglot checks to it. Every language gets a
# OrthographiesCheck and a NoOrphanedMarksInOrthographiesCheck, and then any
# additional checks defined in the language definition file.
from pathlib import Path
import yaml

from gflanguages import LoadLanguages, LoadScripts
from strictyaml import YAMLValidationError
from strictyaml import load as strictyaml_load
from google.protobuf.json_format import MessageToDict

from .checks import schemas, checks_map
from .checks.no_orphaned_marks import NoOrphanedMarksInOrthographiesCheck
from .checks.orthographies import OrthographiesCheck
from .providers import PROVIDERS

gflangs = LoadLanguages()
gfscripts = LoadScripts()

definitions_directory = Path(__file__).parent / "languages"

definition_cache = {}

def load_shaperglot_definition(language, validate=False):
    if language in definition_cache:
        return definition_cache[language]
    definition_file = definitions_directory / (language + ".yaml")
    if not definition_file.is_file():
        return []
    with open(definition_file, encoding="utf-8") as fh:
        if validate:
            data = strictyaml_load(fh.read())
        else:
            data = yaml.safe_load(fh.read())
    check_objects = []
    for ix, check in enumerate(data):
        if validate:
            if check["check"] not in schemas:
                raise ValueError(
                    f"Language definition file for {language} invalid; "
                    f"unknown check type {check['check']} in check {ix} "
                    f"- choose from { ', '.join(schemas.keys()) }"
                )
            try:
                check.revalidate(schemas[check["check"]])
            except YAMLValidationError as e:
                raise ValueError(
                    f"Language definition file for {language} invalid; "
                    f"parser error in check {ix}: {e}"
                ) from e
            check = check.data
        # This turns a { "check": "foobar" } into a FoobarCheck({"check": "foobar"})
        check_objects.append(checks_map[check["check"]](check))
    definition_cache[language] = check_objects
    return check_objects


class Languages:
    loaded = {}

    def __contains__(self, item):
        return item in gflangs

    def keys(self):
        return gflangs.keys()

    def disambiguate(self, lang):
        maybe_keys = [k for k in gflangs if k.lower().startswith(lang.lower() + "_")]
        if maybe_keys:
            return maybe_keys
        maybe_keys = [
            k for k, v in gflangs.items() if v.name.lower().startswith(lang.lower())
        ]
        return maybe_keys

    def __getitem__(self, item):
        if item in self.loaded:
            return self.loaded[item]
        if item not in gflangs:
            return None
        orig = MessageToDict(gflangs[item])
        orig["full_name"] = (
            orig["name"] + " in the " + gfscripts[orig["script"]].name + " script"
        )
        orig["shaperglot_checks"] = [
            OrthographiesCheck(orig),
            NoOrphanedMarksInOrthographiesCheck(orig)
        ]
        for provider in PROVIDERS:
            provider.fill(orig)
        checks = load_shaperglot_definition(item, validate=False)
        orig["shaperglot_checks"].extend(checks)
        self.loaded[item] = orig
        return orig



from ufo2ft.util import closeGlyphsOverGSUB
from vharfbuzz import Vharfbuzz
from fontTools.ttLib import TTFont
from shaperglot.reporter import Reporter


def flatten(lst):
    return [item for sublist in lst for item in sublist]


class Checker:  # pylint: disable=too-few-public-methods
    """A class that creates a checking context for a font. We can then call
    `.check` on this context to run all the checks for a given language, returning
    a Reporter object with the results.
    """
    def __init__(self, fontfile):
        self.vharfbuzz = Vharfbuzz(fontfile)
        self.ttfont = TTFont(fontfile)
        self.glyphorder = self.ttfont.getGlyphOrder()
        self.cmap = self.ttfont["cmap"].getBestCmap()
        self.reversed_cmap = self.ttfont["cmap"].buildReversed()
        self.full_reversed_cmap = None
        self.results = None
        self.lang = None
        self.cache = {"can_shape": {}}

    def _build_full_reversed_cmap(self):
        gsub = self.ttfont.get("GSUB")
        self.full_reversed_cmap = {k: list(v)[0] for k,v in self.reversed_cmap.items()}
        if not gsub:
            return
        if len(self.full_reversed_cmap) > 5_000:
            # Some kind of CJK monstrosity, give up
            return
        for codepoint, glyph in self.cmap.items():
            glyphs = set([glyph])
            if gsub:
                closeGlyphsOverGSUB(gsub, glyphs)
            for new_glyph in glyphs:
                self.full_reversed_cmap[new_glyph] = codepoint

    def codepoint_for(self, glyphname):
        if glyphname in self.reversed_cmap:
            return list(self.reversed_cmap[glyphname])[0]
        if not self.full_reversed_cmap:
            self._build_full_reversed_cmap()
        return self.full_reversed_cmap.get(glyphname, 0)

    def check(self, lang, fail_fast=False):
        self.results = Reporter()
        self.lang = lang
        for check_object in self.lang.get("shaperglot_checks", []):
            skip_reason = check_object.should_skip(self)
            if skip_reason:
                self.results.skip(check_name=check_object.name, message=skip_reason)
                continue
            check_object.execute(self)
            if fail_fast and self.results.fails:
                return self.results
        return self.results



from strictyaml import Seq, Map

from .orthographies import OrthographiesCheck
from .shaping_differs import ShapingDiffersCheck
from .no_orphaned_marks import NoOrphanedMarksCheck
from .unencoded_variants import UnencodedVariantsCheck

checks = [OrthographiesCheck, ShapingDiffersCheck, NoOrphanedMarksCheck, UnencodedVariantsCheck]
checks_map = {cls.name: cls for cls in checks}
schemas = {cls.name: cls.schema for cls in checks}



from enum import Enum
from collections.abc import Sequence
from dataclasses import dataclass, field

from termcolor import colored

class Result(Enum):
    PASS = colored("PASS", "green")
    WARN = colored("WARN", "yellow")
    FAIL = colored("FAIL", "red")
    SKIP = colored("SKIP", "blue")


@dataclass(repr=False)
class Message():
    result: Result
    check_name: str
    message: str
    result_code: str = "ok"
    context: dict = field(default_factory=dict)

    def __repr__(self):
        return self.result.value + ": " + self.message


class Reporter(Sequence):
    def __init__(self):
        self.results = []

    def __getitem__(self, index):
        return self.results[index]

    def __len__(self):
        return len(self.results)

    def okay(self, **kwargs):
        self.results.append(Message(result=Result.PASS, **kwargs))

    def warn(self, **kwargs):
        self.results.append(Message(result=Result.WARN, **kwargs))

    def fail(self, **kwargs):
        self.results.append(Message(result=Result.FAIL, **kwargs))

    def skip(self, **kwargs):
        self.results.append(Message(result=Result.SKIP, **kwargs))

    @property
    def is_unknown(self):
        return not self.passes and not self.fails

    @property
    def is_success(self):
        return len(self.passes) > 0 and not self.fails

    @property
    def passes(self):
        return [x for x in self.results if x.result == Result.PASS]

    @property
    def fails(self):
        return [x for x in self.results if x.result == Result.FAIL]

    @property
    def warns(self):
        return [x for x in self.results if x.result == Result.WARN]



import unicodedata

from shaperglot.checks.no_orphaned_marks import NoOrphanedMarksCheck
from shaperglot.checks.shaping_differs import ShapingDiffersCheck
from shaperglot.checks.unencoded_variants import UnencodedVariantsCheck

isoconv = {
    "aar": "aa",
    "afr": "af",
    "aka": "ak",
    "amh": "am",
    "bam": "bm",
    "ewe": "ee",
    "ful": "ff",
    "hau": "ha",
    "her": "hz",
    "ibo": "ig",
    "kik": "ki",
    "kin": "rw",
    "kon": "kg",
    "kua": "kj",
    "lin": "ln",
    "lub": "lu",
    "lug": "lg",
    "mlg": "mg",
    "nbl": "nr",
    "nde": "nd",
    "ndo": "ng",
    "nya": "ny",
    "orm": "om",
    "run": "rn",
    "sag": "sg",
    "sna": "sn",
    "som": "so",
    "sot": "st",
    "ssw": "ss",
    "swa": "sw",
    "tir": "ti",
    "tsn": "tn",
    "tso": "ts",
    "twi": "tw",
    "ven": "ve",
    "wol": "wo",
    "xho": "xh",
    "yor": "yo",
}

afr_tags = [
    "aaa", "aab", "aal", "aao", "aar", "aas", "aba", "abb", "abi",
    "abm", "abn", "abo", "abr", "abu", "acb", "acd", "ach", "acp", 
    "acq", "acw", "acx", "acz", "ada", "add", "ade", "adh", "adj", 
    "adq", "adu", "aeb", "aec", "ael", "afb", "afe", "afn", "afo", 
    "afr", "aft", "afu", "agb", "agc", "agh", "agj", "agq", "ags", 
    "aha", "ahg", "ahi", "ahl", "ahm", "ahn", "ahp", "ahs", "aik", 
    "aiw", "aiy", "aja", "ajg", "aju", "ajw", "aka", "akd", "akf", 
    "akp", "aks", "aku", "akw", "ala", "ald", "alf", "alw", "alz", 
    "amb", "amf", "amh", "amj", "amo", "anc", "anf", "ank", "ann", 
    "anu", "anv", "anw", "any", "apd", "aqd", "aqg", "arb", "arq", 
    "arv", "ary", "arz", "asa", "asg", "asj", "ass", "asv", "atg", 
    "ati", "ato", "atu", "aug", "auh", "auj", "aum", "avi", "avn", 
    "avu", "awc", "awn", "awo", "axk", "ayb", "aye", "ayg", "ayi", 
    "ayk", "ayl", "ayu", "azo", "bab", "baf", "bag", "bam", "bas", 
    "bau", "bav", "baw", "bax", "bba", "bbe", "bbg", "bbi", "bbj", 
    "bbk", "bbm", "bbo", "bbp", "bbq", "bbs", "bbt", "bbu", "bbw", 
    "bbx", "bby", "bcb", "bce", "bcg", "bci", "bcn", "bcp", "bcq", 
    "bcs", "bct", "bcv", "bcw", "bcy", "bcz", "bda", "bde", "bdh", 
    "bdi", "bdj", "bdm", "bdn", "bdo", "bdp", "bds", "bdt", "bdu", 
    "beb", "bec", "beh", "bej", "bem", "beq", "bes", "bet", "bev", 
    "bex", "bez", "bfa", "bfd", "bff", "bfj", "bfl", "bfm", "bfo", 
    "bfp", "bga", "bgf", "bgj", "bgo", "bgu", "bhs", "bhy", "bib", 
    "bid", "bif", "bij", "bil", "bim", "bin", "bip", "biv", "biw", 
    "biz", "bja", "bjg", "bji", "bjo", "bjt", "bju", "bjv", "bjw", 
    "bka", "bkc", "bkf", "bkg", "bkh", "bkj", "bkm", "bko", "bkp", 
    "bkt", "bkv", "bkw", "bky", "ble", "blh", "bli", "blm", "blo", 
    "blv", "bly", "bma", "bmb", "bmd", "bme", "bmf", "bmg", "bmi", 
    "bml", "bmo", "bmq", "bms", "bmv", "bmw", "bng", "bni", "bnl", 
    "bnm", "bnx", "bnz", "bob", "boe", "bof", "boh", "bok", "bol", 
    "bom", "boo", "bot", "bou", "bov", "box", "boy", "boz", "bpd", 
    "bpj", "bqa", "bqc", "bqd", "bqf", "bqg", "bqj", "bqk", "bqm", 
    "bqo", "bqp", "bqt", "bqu", "bqv", "bqw", "bqx", "bqz", "brf", 
    "bri", "brk", "brl", "brm", "brt", "bsc", "bse", "bsf", "bsi", 
    "bsj", "bsl", "bso", "bsp", "bsq", "bsr", "bss", "bst", "bsv", 
    "bsw", "bsx", "bta", "btc", "bte", "btf", "btg", "btt", "btu", 
    "bub", "bud", "buf", "bui", "buj", "bum", "bun", "bus", "buu", 
    "buw", "bux", "buy", "buz", "bva", "bvb", "bvf", "bvg", "bvh", 
    "bvi", "bvj", "bvm", "bvo", "bvq", "bvw", "bvx", "bwc", "bwg", 
    "bwh", "bwj", "bwl", "bwo", "bwq", "bwr", "bws", "bwt", "bwu", 
    "bww", "bwy", "bwz", "bxb", "bxc", "bxe", "bxg", "bxk", "bxl", 
    "bxp", "bxq", "bxs", "bxv", "bxw", "byb", "byc", "byf", "byg", 
    "byi", "byj", "byn", "byp", "bys", "byt", "byv", "bza", "bze", 
    "bzm", "bzo", "bzv", "bzw", "bzx", "bzy", "bzz", "cae", "cbj", 
    "cbo", "cbq", "cce", "ccg", "cch", "ccj", "cdr", "cen", "cet", 
    "cfa", "cfd", "cfg", "cgg", "chw", "cib", "cie", "cjk", "ckl", 
    "cko", "ckq", "ckx", "cky", "cla", "cli", "cll", "cme", "cmt", 
    "cnu", "coh", "cop", "cou", "cpn", "cpo", "cra", "cry", "csk", 
    "cug", "cuh", "cuv", "cwa", "cwb", "cwe", "cwt", "daa", "dae", 
    "dag", "dai", "daj", "dal", "dam", "das", "dau", "dav", "dba", 
    "dbb", "dbd", "dbg", "dbi", "dbm", "dbo", "dbp", "dbq", "dbr", 
    "dbt", "dbu", "dbv", "dbw", "ddd", "dde", "ddn", "dds", "dec", 
    "dee", "deg", "dek", "deq", "dez", "dga", "dgb", "dgd", "dgh", 
    "dgi", "dgk", "dgl", "dgs", "dhm", "dhs", "dib", "dic", "did", 
    "dig", "dii", "dik", "dil", "dim", "dio", "dip", "dir", "diu", 
    "diw", "diz", "djc", "dje", "djm", "dks", "dkx", "dlk", "dma", 
    "dmb", "dme", "dmm", "dmo", "dmx", "dne", "dnj", "dnn", "dno", 
    "doe", "doh", "doo", "dop", "dos", "dot", "dov", "dow", "dox", 
    "doy", "doz", "drb", "dri", "drs", "dsh", "dsi", "dsq", "dti", 
    "dtk", "dtm", "dtn", "dto", "dts", "dtt", "dtu", "dua", "dug", 
    "dur", "dux", "duz", "dwa", "dwr", "dya", "dyi", "dym", "dyo", 
    "dyu", "dza", "dzg", "dzn", "ebg", "ebo", "ebr", "ebu", "efa", 
    "efe", "efi", "ega", "ego", "ehu", "eja", "eka", "eke", "eki", 
    "ekm", "eko", "ekp", "ekr", "elh", "eli", "elm", "elo", "ema", 
    "emk", "emn", "enb", "enn", "env", "enw", "eot", "epi", "erh", 
    "etb", "eto", "ets", "etu", "etx", "evh", "ewe", "ewo", "eyo", 
    "eza", "eze", "fah", "fak", "fal", "fam", "fan", "fap", "fer", 
    "ffm", "fgr", "fia", "fie", "fip", "fir", "fkk", "fli", "fll", 
    "flr", "fmp", "fni", "fod", "fom", "fon", "fub", "fuc", "fue", 
    "fuf", "fuh", "fui", "fuj", "fum", "fuq", "fuu", "fuv", "fvr", 
    "fwe", "gaa", "gab", "gax", "gaz", "gbg", "gbh", "gbn", "gbo", 
    "gbp", "gbq", "gbr", "gbs", "gbv", "gbx", "gby", "gde", "gdf", 
    "gdi", "gdk", "gdl", "gdm", "gdu", "gea", "gec", "ged", "geg", 
    "gej", "gek", "gel", "geq", "gev", "gew", "gex", "gey", "gez", 
    "ggb", "ggu", "gha", "ghl", "gho", "gic", "gid", "gie", "gis", 
    "gix", "giz", "gji", "gjn", "gke", "gkn", "gkp", "gku", "glc", 
    "glj", "glo", "glr", "glu", "glw", "gly", "gmd", "gmm", "gmn", 
    "gmv", "gmx", "gmz", "gna", "gnd", "gne", "gng", "gnh", "gnj", 
    "gnk", "gnz", "goa", "god", "gof", "gog", "gol", "gou", "gow", 
    "gox", "goy", "gpa", "gqa", "gqr", "grd", "grh", "grj", "grr", 
    "gru", "grv", "gry", "gsl", "gso", "gua", "gud", "guk", "gur", 
    "guw", "gux", "guz", "gvl", "gvm", "gwa", "gwb", "gwd", "gwe", 
    "gwg", "gwj", "gwn", "gwr", "gwx", "gxx", "gya", "gye", "gyg", 
    "gyi", "gyl", "gza", "hae", "hag", "han", "haq", "har", "hau", 
    "hav", "hay", "hba", "hbb", "hbn", "hdy", "hed", "heh", "hem", 
    "her", "hgm", "hhr", "hia", "hig", "hij", "hio", "hka", "hke", 
    "hmb", "hna", "hnh", "hod", "hoe", "hol", "hom", "hoo", "hor", 
    "hoz", "hts", "huc", "hum", "hwa", "hwo", "hya", "ibb", "ibe", 
    "ibm", "ibn", "ibo", "ibr", "iby", "ica", "ich", "ida", "idc", 
    "idd", "ide", "idr", "ids", "idu", "ife", "ifm", "igb", "ige", 
    "igl", "igw", "ihi", "ijc", "ije", "ijj", "ijn", "ijs", "iki", 
    "ikk", "ikl", "iko", "ikp", "ikv", "ikw", "ikx", "ikz", "ilb", 
    "ilv", "ior", "iqw", "iri", "irk", "ish", "isi", "isn", "iso", 
    "isu", "itm", "its", "itw", "iya", "iyo", "iyx", "izr", "izz", 
    "jab", "jad", "jaf", "jbe", "jbn", "jbu", "jek", "jen", "jer", 
    "jeu", "jgb", "jgk", "jgo", "jia", "jib", "jid", "jie", "jii", 
    "jim", "jit", "jjr", "jku", "jle", "jmb", "jmc", "jmi", "jmr", 
    "jms", "jni", "jnj", "job", "jod", "jow", "jrr", "jrt", "jub", 
    "jud", "juh", "juk", "jum", "juo", "juu", "juw", "jwi", "jyy", 
    "kab", "kad", "kah", "kai", "kaj", "kam", "kao", "kbj", "kbl", 
    "kbn", "kbo", "kbp", "kbr", "kbs", "kby", "kbz", "kcc", "kce", 
    "kcf", "kcg", "kch", "kci", "kcj", "kck", "kcm", "kcp", "kcq", 
    "kcr", "kcs", "kcu", "kcv", "kcw", "kcx", "kcy", "kcz", "kdc", 
    "kde", "kdg", "kdh", "kdi", "kdj", "kdl", "kdm", "kdn", "kdp", 
    "kdu", "kdx", "kdz", "keb", "kec", "ked", "kef", "keg", "kel", 
    "ken", "keo", "ker", "kes", "keu", "kez", "kfl", "kfn", "kfo", 
    "kfz", "kga", "kgo", "kgt", "khj", "khq", "khu", "khx", "khy", 
    "kia", "kib", "kid", "kie", "kik", "kil", "kin", "kiv", "kiz", 
    "kka", "kkd", "kke", "kki", "kkj", "kkm", "kko", "kkq", "kkr", 
    "kks", "kku", "kkw", "klc", "klf", "klk", "klo", "klu", "kma", 
    "kmb", "kme", "kmi", "kmp", "kmq", "kmw", "kmy", "kna", "knc", 
    "knf", "kng", "kni", "knk", "kno", "knp", "knu", "knw", "kny", 
    "knz", "koc", "koe", "kof", "koh", "kon", "koo", "koq", "kot", 
    "kou", "kov", "kow", "kpa", "kph", "kpk", "kpl", "kpo", "kpz", 
    "kqg", "kqh", "kqk", "kqm", "kqn", "kqo", "kqp", "kqs", "kqu", 
    "kqx", "kqy", "kqz", "krh", "krn", "krp", "krs", "krt", "krw", 
    "krx", "ksa", "ksb", "ksf", "ksm", "kso", "ksp", "ksq", "kss", 
    "kst", "ksv", "ktb", "ktc", "ktf", "kth", "ktj", "kty", "ktz", 
    "kua", "kub", "kug", "kuh", "kuj", "kul", "kun", "kus", "kuw", 
    "kvf", "kvi", "kvj", "kvm", "kwb", "kwc", "kwg", "kwl", "kwm", 
    "kwn", "kwp", "kws", "kwu", "kwv", "kwy", "kwz", "kxb", "kxc", 
    "kxh", "kxj", "kxx", "kya", "kye", "kyf", "kym", "kyq", "kza", 
    "kzc", "kzn", "kzo", "kzr", "kzy", "laf", "lag", "lai", "laj", 
    "lak", "lal", "lam", "lan", "lap", "lar", "las", "lbi", "lch", 
    "lda", "ldb", "ldd", "ldg", "ldh", "ldi", "ldj", "ldk", "ldl", 
    "ldm", "ldo", "ldp", "ldq", "lea", "leb", "led", "lee", "lef", 
    "leh", "lej", "lel", "lem", "leo", "les", "lfa", "lgg", "lgm", 
    "lgn", "lgq", "lgz", "lia", "lie", "lig", "lik", "lin", "lip", 
    "liq", "liu", "liy", "liz", "lkb", "lke", "lko", "lkr", "lks", 
    "lky", "lla", "llb", "llc", "lli", "lln", "lma", "lmd", "lme", 
    "lmi", "lmp", "lmx", "lna", "lnb", "lnl", "lno", "lns", "lnu", 
    "lnz", "lob", "lof", "log", "loh", "loi", "lok", "lol", "lom", 
    "lon", "loo", "lop", "loq", "lor", "lot", "loz", "lpx", "lri", 
    "lrm", "lro", "lse", "lsm", "lth", "lto", "lts", "lua", "lub", 
    "luc", "lue", "lug", "luj", "lul", "lum", "lun", "luo", "lup", 
    "luq", "luw", "lwa", "lwg", "lwo", "lyn", "mae", "maf", "mas", 
    "maw", "mbm", "mbo", "mbu", "mbv", "mcj", "mck", "mcn", "mcp", 
    "mcs", "mct", "mcu", "mcw", "mcx", "mda", "mdd", "mde", "mdg", 
    "mdi", "mdj", "mdk", "mdm", "mdn", "mdp", "mdq", "mdt", "mdu", 
    "mdw", "mdx", "mdy", "mea", "mei", "men", "meq", "mer", "mes", 
    "mev", "mew", "mey", "mfc", "mfd", "mff", "mfg", "mfh", "mfi", 
    "mfj", "mfk", "mfl", "mfm", "mfn", "mfo", "mfq", "mfu", "mfv", 
    "mfx", "mfz", "mgb", "mgc", "mgd", "mge", "mgg", "mgh", "mgi", 
    "mgj", "mgn", "mgo", "mgq", "mgr", "mgs", "mgv", "mgw", "mgy", 
    "mgz", "mhb", "mhd", "mhi", "mhk", "mhm", "mho", "mhw", "mif", 
    "mij", "mje", "mjh", "mjs", "mka", "mkf", "mkk", "mkl", "mko", 
    "mku", "mlb", "mlg", "mlj", "mlk", "mlo", "mlq", "mlr", "mls", 
    "mlt", "mlw", "mma", "mmf", "mmu", "mmy", "mmz", "mne", "mnf", 
    "mnh", "mnk", "mny", "moa", "moi", "moj", "mor", "mos", "mou", 
    "mow", "moy", "moz", "mpa", "mpe", "mpg", "mpi", "mpk", "mqb", 
    "mql", "mqu", "mrt", "mru", "msc", "mse", "msj", "msv", "msw", 
    "mtb", "mtk", "mtl", "mua", "mub", "muc", "mug", "muh", "muj", 
    "muo", "mur", "muu", "muy", "muz", "mvh", "mvu", "mvw", "mvz", 
    "mwe", "mwk", "mwm", "mwn", "mws", "mwu", "mwz", "mxc", "mxf", 
    "mxg", "mxh", "mxl", "mxo", "mxu", "mxx", "myb", "myc", "mye", 
    "myf", "myg", "myj", "myk", "mym", "myo", "mys", "myx", "mzb", 
    "mzd", "mzj", "mzk", "mzm", "mzv", "mzw", "naj", "naq", "nar", 
    "nat", "naw", "nba", "nbb", "nbd", "nbh", "nbl", "nbm", "nbo", 
    "nbp", "nbr", "nbv", "nbw", "ncr", "ncu", "nda", "ndb", "ndc", 
    "ndd", "nde", "ndg", "ndh", "ndi", "ndj", "ndk", "ndl", "ndm", 
    "ndn", "ndo", "ndp", "ndq", "ndr", "ndt", "ndu", "ndv", "ndw", 
    "ndy", "ndz", "neb", "ned", "ney", "nfd", "nfr", "nfu", "nga", 
    "ngb", "ngc", "ngd", "nge", "ngg", "ngh", "ngi", "ngj", "ngl", 
    "ngn", "ngo", "ngp", "ngq", "ngs", "ngv", "ngw", "ngx", "ngy", 
    "ngz", "nhb", "nhr", "nhu", "nie", "nih", "nim", "nin", "niq", 
    "nix", "niy", "nja", "njd", "njj", "njl", "njr", "njx", "njy", 
    "nka", "nkc", "nkn", "nko", "nkq", "nkt", "nku", "nkv", "nkw", 
    "nkx", "nkz", "nla", "nle", "nlj", "nlo", "nlu", "nmc", "nmd", 
    "nmg", "nmi", "nmj", "nml", "nmn", "nmq", "nmr", "nmz", "nnb", 
    "nnc", "nne", "nnh", "nnj", "nnn", "nnq", "nnu", "nnw", "nnz", 
    "noq", "now", "noy", "noz", "nqg", "nqk", "nql", "nqo", "nra", 
    "nrb", "nsc", "nse", "nsg", "nsh", "nso", "nsx", "nte", "nti", 
    "ntk", "ntm", "nto", "ntr", "nue", "nuh", "nui", "nuj", "nup", 
    "nus", "nuu", "nuv", "nvo", "nwb", "nwe", "nwm", "nxd", "nxi", 
    "nxo", "nya", "nyb", "nyc", "nyd", "nye", "nyf", "nyg", "nyi", 
    "nyj", "nyk", "nym", "nyn", "nyo", "nyp", "nyr", "nyu", "nyy", 
    "nza", "nzb", "nzd", "nzi", "nzk", "nzu", "nzy", "nzz", "obl", 
    "obu", "oda", "odu", "ofu", "ogb", "ogc", "ogg", "ogo", "ogu", 
    "okb", "okd", "oke", "oki", "okr", "oks", "oku", "okx", "old", 
    "olm", "olu", "omi", "oml", "omt", "opa", "orc", "org", "orm", 
    "orr", "orx", "oso", "ost", "otr", "oua", "oub", "oyd", "ozm", 
    "pae", "pai", "pbi", "pbl", "pbn", "pbo", "pbp", "pbr", "pcn", 
    "pcw", "pem", "pfe", "pgs", "phm", "pic", "pil", "pip", "piw", 
    "piy", "pkb", "pko", "plj", "plr", "pmb", "pmm", "pmn", "pnd", 
    "png", "pnl", "pnq", "pny", "pnz", "pof", "poy", "ppp", "pqa", 
    "pug", "puu", "pwb", "pye", "pym", "rag", "ras", "rax", "reg", 
    "rel", "rer", "res", "rif", "rim", "rin", "rkm", "rnd", "rng", 
    "rnw", "rod", "rof", "rou", "rub", "ruc", "ruf", "rui", "ruk", 
    "run", "ruy", "ruz", "rwk", "rwm", "rzh", "saa", "sad", "saf", 
    "sag", "sak", "saq", "sav", "say", "sba", "sbd", "sbf", "sbj", 
    "sbk", "sbm", "sbp", "sbs", "sbw", "sby", "sbz", "scv", "scw", 
    "sde", "sdj", "sds", "seb", "sef", "seg", "seh", "sen", "sep", 
    "seq", "ses", "sev", "sfw", "sgc", "sgi", "sgm", "sgw", "sha", 
    "shc", "she", "shg", "shi", "shj", "shk", "sho", "shq", "shr", 
    "shu", "shw", "shy", "shz", "sid", "sie", "sif", "sig", "sil", 
    "sir", "siz", "sjg", "sjs", "skq", "skt", "sld", "slx", "smd", 
    "smx", "sna", "snf", "sng", "snk", "snm", "snq", "snw", "soc", 
    "sod", "soe", "soh", "sok", "som", "soo", "sop", "sor", "sos", 
    "sot", "sox", "soy", "soz", "spp", "spy", "sqa", "sqh", "sqm", 
    "srr", "ssc", "ssl", "ssn", "ssw", "ssy", "stj", "stv", "sub", 
    "suj", "suk", "suq", "sur", "sus", "suw", "swa", "swb", "swc", 
    "swf", "swh", "swj", "swk", "swn", "swq", "swy", "sxb", "sxe", 
    "sxs", "sxw", "syc", "syi", "syk", "sym", "sys", "syx", "sze", 
    "szg", "szv", "tag", "tak", "tal", "tan", "tap", "taq", "tax", 
    "taz", "tbi", "tbm", "tbr", "tbt", "tbz", "tcc", "tcd", "tck", 
    "tda", "tde", "tdk", "tdl", "tdo", "tdq", "tdv", "tec", "ted", 
    "teg", "tek", "tem", "teo", "teq", "teu", "tex", "tey", "tez", 
    "tfi", "tga", "tgd", "tgw", "tgy", "thk", "thu", "thv", "thy", 
    "thz", "tia", "tic", "tig", "tii", "tik", "tiq", "tir", "tiv", 
    "tja", "tjn", "tjo", "tke", "tkq", "tlj", "tll", "tlo", "tma", 
    "tmc", "tms", "tmv", "tng", "tnr", "tny", "tod", "tog", "toh", 
    "toi", "toq", "tor", "toz", "tpm", "tqq", "tqr", "trj", "tsa", 
    "tsb", "tsc", "tsh", "tsn", "tso", "tsp", "tst", "tsv", "tsw", 
    "ttb", "ttf", "ttj", "ttl", "ttq", "ttr", "tug", "tui", "tul", 
    "tum", "tuq", "tuv", "tuy", "tuz", "tvd", "tvs", "tvu", "twi", 
    "twl", "twn", "two", "twq", "twx", "txj", "tye", "tyi", "tyu", 
    "tyx", "tzm", "uba", "ubi", "uda", "udl", "udu", "uha", "uiv", 
    "uji", "ukh", "ukp", "ukq", "uku", "ukw", "ula", "ulb", "umb", 
    "umm", "une", "urh", "usk", "uss", "uta", "uth", "utr", "uya", 
    "vae", "vag", "vai", "vaj", "vau", "vem", "ven", "ver", "vid", 
    "vif", "vig", "vin", "vit", "vki", "vkj", "vmk", "vmr", "vmw", 
    "vor", "vum", "vun", "vut", "wal", "wan", "wav", "wbf", "wbh", 
    "wbi", "wbj", "wci", "wdd", "wec", "weh", "wem", "wib", "wja", 
    "wji", "wka", "wlc", "wle", "wll", "wlx", "wma", "wmw", "wni", 
    "wob", "wof", "wok", "wol", "wom", "won", "woy", "wrn", "wss", 
    "wti", "wud", "wum", "wun", "wwa", "www", "xab", "xam", "xan", 
    "xdo", "xed", "xeg", "xel", "xho", "xii", "xkb", "xkg", "xkt", 
    "xku", "xkv", "xma", "xmb", "xmc", "xmd", "xmg", "xmj", "xnz", 
    "xoc", "xog", "xom", "xon", "xpe", "xrb", "xsh", "xsj", "xsm", 
    "xsn", "xsq", "xtc", "xuo", "xuu", "xwe", "xwg", "xwl", "yaf", 
    "yaj", "yal", "yam", "yao", "yas", "yat", "yav", "yax", "yay", 
    "yaz", "yba", "ybb", "ybj", "ybl", "yei", "yel", "yer", "yes", 
    "yey", "yko", "yky", "ymg", "ymk", "ymm", "yng", "ynq", "yns", 
    "yom", "yor", "yot", "yre", "yul", "yun", "zag", "zah", "zaj", 
    "zak", "zay", "zaz", "zdj", "zen", "zga", "zgh", "zhi", "zhw", 
    "zil", "zim", "zin", "zir", "ziw", "ziz", "zmb", "zmf", "zmn", 
    "zmo", "zmp", "zmq", "zms", "zmw", "zmx", "zmz", "zna", "zne", 
    "zns", "zrn", "zua", "zul", "zuy", "zwa"
]

unencoded_variants = {
    "tod_Latn": ["ʋ", "Ʋ"],
    "xpe_Latn": ["Ɓ"],
    "lom_Latn": ["Ɓ"],
    "dnj_Latn": ["Ɓ"],
    "gaa_Latn": ["Ʃ", "Ʒ"],
}

available_languages = {
    isoconv.get(tag, tag)+"_Latn": tag for tag in afr_tags
}

class AfricanLatinProvider:
    @classmethod
    def fill(cls, language):
        if language["script"] != "Latn":
            return
        if language["id"] not in available_languages:
            return
        
        exemplar_chars = language.get("exemplarChars", {})
        bases = exemplar_chars.get("base", "").split() or []
        auxiliary = exemplar_chars.get("auxiliary", "").split() or []

        language["shaperglot_checks"].extend(
            cls.build_no_orphaned_marks(bases, auxiliary)
        )
        language["shaperglot_checks"].extend(
            cls.build_unencoded_variants(language, bases, auxiliary)
        )
        language["shaperglot_checks"].extend(
            cls.build_sd_smcp(bases, auxiliary)
        )

    @classmethod
    def build_no_orphaned_marks(cls, bases, auxiliary):
        basemarks = []
        for base in bases:
            if len(base) > 1:
                unpack = base.replace("{", "").replace("}", "")
                for char in unpack:
                    if unicodedata.combining(char):
                        basemarks.append(unpack)
        for aux in auxiliary:
            if len(aux) > 1:
                unpack = aux.replace("{", "").replace("}", "")
                for char in unpack:
                    if unicodedata.combining(char):
                        basemarks.append(unpack)
        basemark_string = "".join(basemarks)
        if not basemark_string:
            return
        yield NoOrphanedMarksCheck(
            {
                "check": "no_orphaned_marks",
                "input": {"text": basemark_string}
            }
        )
        yield NoOrphanedMarksCheck(
            {
                "check": "no_orphaned_marks",
                "input": {"text": basemark_string, "features": {"smcp": True}},
                "conditions": {"features": ['smcp']},
            }
        )

    @classmethod
    def build_unencoded_variants(cls, language, bases, auxiliary):
        if "Ŋ" in bases or 'Ŋ' in auxiliary:
            yield UnencodedVariantsCheck({
                "check": "unencoded_variants", 
                "input": {"text": 'Ŋ'}
            })
        for char in unencoded_variants.get(language["id"], []):
            yield UnencodedVariantsCheck({
                "check": "unencoded_variants", 
                "input": {"text": char}
            })

    @classmethod
    def build_sd_smcp(cls, bases, auxiliary):
        for letter in bases + auxiliary:
            if len(letter) == 1 and unicodedata.category(letter) == 'Ll':
                yield ShapingDiffersCheck({
                    'check': 'shaping_differs',
                    'inputs': [{'text': letter}, {'text': letter, 'features': {'smcp': True}}],
                    'conditions': {'features': ['smcp']},
                    'rationale': "Requires Small-cap: " + letter,
                })



from strictyaml import FixedSeq, Map, Int, Optional
from num2words import num2words

from .common import shaping_input_schema, ShaperglotCheck, check_schema

cluster_schema = Map({Optional("cluster"): Int(), "glyph": Int()})


class ShapingDiffersCheck(ShaperglotCheck):
    name = "shaping_differs"
    schema = check_schema(
        {
            "inputs": FixedSeq([shaping_input_schema, shaping_input_schema]),
            Optional("differs"): FixedSeq([cluster_schema, cluster_schema]),
        }
    )

    def describe(self):
        cluster_desc = []
        if "differs" in self.definition:
            for differs in self.definition["differs"]:
                desc = f"the {num2words(1+int(differs['glyph']), to='ordinal')} glyph"
                if "cluster" in differs:
                    desc += f" of the {num2words(1+int(differs['cluster']), to='ordinal')} cluster"
                cluster_desc.append(desc)
            result = (
                f"{cluster_desc[0]} of the first output is different to "
                f"{cluster_desc[1]} of the second output."
            )
        else:
            result = "the outputs differ."
        full_result = (
            f"that, when {self.inputs[0].describe()}, and then {self.inputs[1].describe()}, "
        ) + result
        if str(self.definition["rationale"]):
            full_result += f" This is because {self.definition['rationale']}."
        return full_result

    def execute(self, checker):
        buffers = [i.shape(checker) for i in self.inputs]
        if "differs" not in self.definition:
            # Any difference is OK
            serialized_buf1 = checker.vharfbuzz.serialize_buf(buffers[0])
            serialized_buf2 = checker.vharfbuzz.serialize_buf(buffers[1])
            if serialized_buf1 != serialized_buf2:
                checker.results.okay(
                    check_name="shaping-differs",
                    message=f"{self.definition['rationale']}",
                )
            else:
                checker.results.fail(
                    check_name="shaping-differs",
                    result_code="shaping-did-not-differ",
                    message=f"{self.definition['rationale']}"
                    + "; both buffers returned "
                    + serialized_buf1,
                    context={
                        "input1": self.inputs[0].check_yaml,
                        "input2": self.inputs[0].check_yaml,
                    },
                )
            return
        # We are looking for a specific difference
        glyphs = []
        for differs, buffer in zip(self.definition["differs"], buffers):
            buffer = list(zip(buffer.glyph_infos, buffer.glyph_positions))
            if "cluster" in differs:
                buffer = [x for x in buffer if x[0].cluster == int(differs["cluster"])]
            glyph_ix = int(differs["glyph"])
            if len(buffer) - 1 < glyph_ix:
                checker.results.fail(
                    check_name="shaping-differs",
                    result_code="too-few-glyphs",
                    message=f"Test asked for glyph {glyph_ix} but shaper only returned {len(buffer)} glyphs",
                    context={
                        "input1": self.inputs[0].check_yaml,
                        "input2": self.inputs[0].check_yaml,
                    },
                )
                return
            glyphs.append((buffer[glyph_ix][0].codepoint, buffer[glyph_ix][1]))
        if glyphs[0] != glyphs[1]:
            checker.results.okay(
                check_name="shaping-differs",
                result_code="shaping-did-not-differ",
                message=f"{self.definition['rationale']}",
            )
        else:
            checker.results.fail(
                check_name="shaping-differs",
                result_code="shaping-did-not-differ",
                message=f"{self.definition['rationale']}"
                + "; both buffers returned "
                + serialized_buf1,
                context={
                    "input1": self.inputs[0].check_yaml,
                    "input2": self.inputs[0].check_yaml,
                },
            )



import re

from strictyaml import Map

from .common import ShaperglotCheck, and_join


def parse_bases(bases):
    return [x[0] or x[1] for x in re.findall(r"\{([^}]+)\}|(\S+)", bases)]


def can_shape(text, checker):
    if text not in checker.cache["can_shape"]:
        buf = checker.vharfbuzz.shape(text)
        checker.cache["can_shape"][text] = all(
            gi.codepoint != 0 for gi in buf.glyph_infos
        )
    return checker.cache["can_shape"][text]


class OrthographiesCheck(ShaperglotCheck):
    name = "orthographies"
    schema = Map({})

    # pylint: disable=W0231
    def __init__(self, lang):
        exemplar_chars = lang.get("exemplarChars", {})
        marks = exemplar_chars.get("marks", "").replace("◌", "").split() or []
        bases = parse_bases(exemplar_chars.get("base", ""))
        aux = parse_bases(exemplar_chars.get("auxiliary", ""))
        self.all_glyphs = marks + bases
        self.marks = set(marks)
        self.bases = set(bases) - self.marks
        self.aux = set(aux) - self.marks

    def should_skip(self, checker):
        return False

    def describe(self):
        return "that the following glyphs are in the font: " + and_join(
            f"'{g}'" for g in self.all_glyphs
        )

    def execute(self, checker):
        if not self.all_glyphs:
            checker.results.warn(
                check_name="orthographies",
                result_code="no-exemplars",
                message=f"No exemplar glyphs were defined for language {checker.lang['name']}",
            )
            return
        missing = sorted([x for x in self.bases if not can_shape(x, checker)])
        if missing:
            checker.results.fail(
                check_name="orthographies",
                result_code="bases-missing",
                message=f"Some base glyphs were missing: {', '.join(missing)}",
                context={"glyphs": missing},
            )
        else:
            checker.results.okay(
                check_name="orthographies",
                message="All base glyphs were present in the font",
            )

        if not self.marks and not self.aux:
            return

        missing = sorted([x for x in self.marks if not can_shape(x, checker)])
        if missing:
            missing_str = ", ".join([chr(0x25CC) + x for x in missing])
            checker.results.fail(
                check_name="orthographies",
                result_code="marks-missing",
                message=f"Some mark glyphs were missing: {missing_str}",
                context={"glyphs": missing},
            )
        else:
            checker.results.okay(
                check_name="orthographies",
                message="All mark glyphs were present in the font",
            )

        if not self.aux:
            return

        missing = sorted([x for x in self.bases if not can_shape(x, checker)])
        if missing:
            checker.results.warn(
                check_name="orthographies",
                result_code="aux-missing",
                message=f"Some auxiliary glyphs were missing: {', '.join(missing)}",
                context={"glyphs": missing},
            )
        else:
            checker.results.okay(
                check_name="orthographies",
                message="All auxiliary glyphs were present in the font",
            )



from strictyaml import Str, Map, Optional, MapPattern, Bool, Seq

shaping_input_schema = Map(
    {
        "text": Str(),
        Optional("language"): Str(),
        Optional("features"): MapPattern(Str(), Bool()),
    }
)

conditions_schema = Map({Optional("features"): Seq(Str())})


def check_schema(schema):
    base_schema = {
        "check": Str(),
        Optional("rationale"): Str(),
        Optional("conditions"): conditions_schema,
    }
    schema.update(base_schema)
    return Map(schema)


def and_join(lst):
    lst = list(lst)
    if len(lst) == 0:
        return "(nothing here)"
    if len(lst) > 1:
        return ', '.join(lst[:-1]) + ' and ' + lst[-1]
    return lst[0]


class ShapeInput:
    def __init__(self, check_yaml):
        self.text = str(check_yaml["text"])
        self.language = str(check_yaml.get("language", ""))
        if "features" in check_yaml:
            self.features = check_yaml["features"]
        else:
            self.features = {}
        self.check_yaml = check_yaml

    def describe(self):
        result = f"shaping the text '{self.text}' "
        if self.language or self.features:
            result += "while "
        if self.language:
            result += f"setting the language to '{self.language}' "
            if self.features:
                result += "and "
        if self.features:
            result += "turning " + and_join(
                [
                    f"the '{feat}' feature " + (state and "on" or "off")
                    for feat, state in self.features.items()
                ]
            )
        return result.strip()

    def shape(self, checker):
        parameters = {}
        if self.language:
            parameters["language"] = self.language
        if self.features:
            parameters["features"] = self.features
        return checker.vharfbuzz.shape(self.text, parameters)


# pylint: disable=R0903
class ShaperglotCheck:
    def __init__(self, check_yaml):
        self.definition = check_yaml
        if "inputs" in check_yaml:
            self.inputs = [ShapeInput(x) for x in check_yaml["inputs"]]
        if "input" in check_yaml:
            self.input = ShapeInput(check_yaml["input"])

    def should_skip(self, checker):
        conditions = self.definition.get("conditions")
        if conditions is None:
            return False
        if "features" in conditions:
            font_features = set()
            for table in ["GSUB", "GPOS"]:
                if (
                    checker.ttfont.get(table)
                    and checker.ttfont[table].table.FeatureList
                ):
                    for feature_record in checker.ttfont[
                        table
                    ].table.FeatureList.FeatureRecord:
                        font_features.add(feature_record.FeatureTag)
            required = set(str(f) for f in conditions["features"])
            if required - font_features:
                return f"Missing features: {', '.join(required-font_features)}"
        return False



import unicodedata
from youseedee import ucd_data

from shaperglot.checks import NoOrphanedMarksCheck, ShapingDiffersCheck

ZWJ = "\u200D"
MANUALLY_DEFINED_MARKS = {
    'ar_Arab': [
        ["\u064E\u0651", 'FATHA+SHADDA'],
        ["\u064B\u0651", 'FATHATAN+SHADDA'],
        ["\u0650\u0651", 'KASRA+SHADDA'],
        ["\u064D\u0651", 'KASRATAN+SHADDA'],
        ["\u064F\u0651", 'DAMMA+SHADDA'],
        ["\u064C\u0651", 'DAMMATAN+SHADDA'],
    ]
}

def get_joining_type(char):
    return ucd_data(ord(char)).get("Joining_Type", "U")

class ArabicProvider:
    @classmethod
    def fill(cls, language):
        if language["script"] != "Arab":
            return
        if "exemplarChars" not in language or "base" not in language["exemplarChars"]:
            return
        # Assemble list of marks
        check_these_marks = []
        for character in language["exemplarChars"].get("marks", []):
            if unicodedata.category(character) == "Mn":
                check_these_marks.append([character, unicodedata.name(character)])
        if language["id"] in MANUALLY_DEFINED_MARKS:
            check_these_marks.extend(MANUALLY_DEFINED_MARKS[language["id"]])
        for character in language["exemplarChars"]["base"]:
            if unicodedata.category(character).startswith("L"):
                joining_type = get_joining_type(character)
                if joining_type in ("R", "D"):
                    language["shaperglot_checks"].extend(cls.shaping_checks(ZWJ, character, '', 'fina'))
                    language["shaperglot_checks"].extend(
                        cls.mark_checks(ZWJ, character, '', check_these_marks, 'fina')
                    )
                if joining_type == "D":
                    language["shaperglot_checks"].extend(cls.shaping_checks(ZWJ, character, ZWJ, 'medi'))
                    language["shaperglot_checks"].extend(
                        cls.mark_checks(ZWJ, character, ZWJ, check_these_marks, 'medi')
                    )
                    language["shaperglot_checks"].extend(cls.shaping_checks('', character, ZWJ, 'init'))
                    language["shaperglot_checks"].extend(
                        cls.mark_checks('', character, ZWJ, check_these_marks, 'init')
                    )
                language["shaperglot_checks"].extend(
                    cls.mark_checks('', character, '', check_these_marks, 'isol')
                )

    @classmethod
    def mark_checks(cls, pre_context, character, post_context, check_these_marks, position):
        for marks, rationale in check_these_marks:
            yield NoOrphanedMarksCheck({
                'input': {"text": pre_context + character + marks + post_context},
                'rationale': f"{rationale} on top of {unicodedata.name(character)} (.{position})",
            })

    @classmethod
    def shaping_checks(cls, pre_context, character, post_context, position):
        yield ShapingDiffersCheck({
            'inputs': [
                {
                    'text': pre_context + character + post_context,
                    'features': {
                        "init": False,
                        "medi": False,
                        "fina": False,
                    },
                },
                {'text': pre_context + character + post_context},
            ],
            'rationale': f".{position} version of {unicodedata.name(character)}",
        })


from .common import shaping_input_schema, ShaperglotCheck, check_schema


class UnencodedVariantsCheck(ShaperglotCheck):
    name = "unencoded_variants"
    schema = check_schema(
        {
            "input": shaping_input_schema,
        }
    )

    def describe(self):
        return f"that, when {self.input.describe()}, an unencoded variant glyph is substituted used the `locl` feature"

    def execute(self, checker):
        if len(self.input.text) > 1:
            raise ValueError(
                f"Please only pass one codepoint at a time to the unencoded variants check (not '{self.input.text}')"
            )
        self.input.features["locl"] = False
        buffer = self.input.shape(checker)
        glyphname = checker.glyphorder[buffer.glyph_infos[0].codepoint]
        # Are there variant versions of this glyph?
        variants = [
            glyph for glyph in checker.glyphorder if glyph.startswith(glyphname + ".")
        ]
        if not variants:
            checker.results.warn(
                check_name="unencoded-variants",
                result_code="no-variant",
                message="No variant glyphs were found for " + glyphname,
                context={"text": self.input.check_yaml, "glyph": glyphname},
            )
            return
        # Try it again with locl on, set the language to the one we're
        # looking for see if something happens.
        if not self.input.language:
            self.input.language = checker.lang["language"]
        self.input.features["locl"] = True
        buffer2 = self.input.shape(checker)
        glyphname2 = checker.glyphorder[buffer2.glyph_infos[0].codepoint]
        if glyphname2 == glyphname:
            checker.results.fail(
                check_name="unencoded-variants",
                result_code="unchanged-after-locl",
                message=f"The locl feature did not affect {glyphname}",
                context={"text": self.input.check_yaml, "glyph": glyphname},
            )
        else:
            checker.results.okay(
                check_name="unencoded-variants",
                message=f"The locl feature changed {glyphname} to {glyphname2}",
                context={"text": self.input.check_yaml, "glyph": glyphname},
            )
